/*
 ** Jo Sega Saturn Engine
 ** Copyright (c) 2012-2016, Johannes Fetz (johannesfetz@gmail.com)
 ** All rights reserved.
 **
 ** Redistribution and use in source and binary forms, with or without
 ** modification, are permitted provided that the following conditions are met:
 **     * Redistributions of source code must retain the above copyright
 **       notice, this list of conditions and the following disclaimer.
 **     * Redistributions in binary form must reproduce the above copyright
 **       notice, this list of conditions and the following disclaimer in the
 **       documentation and/or other materials provided with the distribution.
 **     * Neither the name of the Johannes Fetz nor the
 **       names of its contributors may be used to endorse or promote products
 **       derived from this software without specific prior written permission.
 **
 ** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 ** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 ** DISCLAIMED. IN NO EVENT SHALL Johannes Fetz BE LIABLE FOR ANY
 ** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 ** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 ** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 ** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 ** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/***************************************************************************************************\
 ** Special Note : Use the collision tool to generate the ZTM (binary) file format.
\***************************************************************************************************/

#include <jo/jo.h>


#define WORK_RAM_LOW 0x00200000

static jo_camera cam;
static float rx, ry;
static int x,y,z;
static bool use_screen_doors;
static bool use_light = true;
static bool use_texture = true;
static unsigned int selected_polygon_index = 0;


/*************
Custom binary file format loading test
*************/
static jo_3d_mesh *jo3dMesh[10];  /*** Static number of elements for now ***/
unsigned short mesh_tot = 0;

void my_draw(void)
{
	jo_printf(12, 1, "*Mesh loading demo*");
	jo_3d_camera_look_at(&cam);
	jo_3d_push_matrix();
	{
		jo_3d_rotate_matrix_rad(rx, ry, 0.0f);
		jo_3d_translate_matrix(x, 100, z);

		/***** Just write your draw list (like for (i=0; i < total_mesh; i++) jo_3d_mesh_draw(jo3dMesh[i]) *******/
		int i = 0;
		for (JO_ZERO(i); i<mesh_tot; i++)
        {
           jo_3d_mesh_draw(jo3dMesh[i]);
        }

		//jo_3d_mesh_draw(&MapTest);
	}
	jo_3d_pop_matrix();

	//jo_printf(0, 23, "Total polygon ( %d )", jo_)
	jo_printf(0, 11, "Total meshes  %d   ", mesh_tot);
	jo_printf(0, 12, "Total quads :  %d  ", jo_3d_get_displayed_polygon_count());
	jo_printf(0, 24, "[PAD] Rotate mesh");
	jo_printf(0, 25, "[L/R] select polygon (index=%d)  ", selected_polygon_index);
	jo_printf(0, 26, "[C] to toggle texture (%s) ", use_texture ? "on" : "off");
	jo_printf(0, 27, "[B] to toggle light (%s) ", use_light ? "on" : "off");
	jo_printf(0, 28, "[A] to toggle screen doors (%s) ", use_screen_doors ? "on" : "off");
}

void apply_effect(void)
{
	jo_3d_set_mesh_screen_doors(jo3dMesh[0], use_screen_doors);
	jo_3d_set_mesh_light(jo3dMesh[0], use_light);
	if (use_texture)
		jo_3d_set_mesh_texture(jo3dMesh[0], 0);
	else
		jo_3d_set_mesh_color(jo3dMesh[0], JO_COLOR_Gray);
	jo_3d_set_mesh_polygon_color(jo3dMesh[0], JO_COLOR_Red, selected_polygon_index);
}

void my_gamepad(void)
{
	if (jo_is_pad1_key_pressed(JO_KEY_UP))
		rx -= 0.01f;
	else if (jo_is_pad1_key_pressed(JO_KEY_DOWN))
		rx += 0.01f;
	else if (jo_is_pad1_key_pressed(JO_KEY_LEFT))
		ry += 0.01f;
	else if (jo_is_pad1_key_pressed(JO_KEY_RIGHT))
		ry -= 0.01f;
	if (jo_is_pad1_key_down(JO_KEY_A))
	{
		use_screen_doors ^= true;
		apply_effect();
	}
	if (jo_is_pad1_key_down(JO_KEY_B))
	{
		use_light ^= true;
		apply_effect();
	}
	if (jo_is_pad1_key_down(JO_KEY_C))
	{
		use_texture ^= true;
		apply_effect();
	}
	if (jo_is_pad1_key_down(JO_KEY_L))
	{
		if (selected_polygon_index > 0)
			--selected_polygon_index;
		else
			selected_polygon_index = jo_3d_get_mesh_polygon_count(jo3dMesh[0]) - 1;
		apply_effect();
	}
	if (jo_is_pad1_key_down(JO_KEY_R))
	{
		++selected_polygon_index;
		if (selected_polygon_index >= jo_3d_get_mesh_polygon_count(jo3dMesh[0]))
			selected_polygon_index = 0;
		apply_effect();
	}
	if (jo_is_pad1_key_pressed(JO_KEY_X)) x--;
	if (jo_is_pad1_key_pressed(JO_KEY_Y)) x++;
}

/**********
Load a binary level (ZTM format) and put it in the jo 3d mesh array
TODO : Fix corruption bug on SSF (works OK on Yabause).
**********/
void * LoadBinaryMAP(void * startAddress)
{

 char * stream;
 void * currentAddress;
 int length;
 mesh_tot = 0;
 stream = jo_fs_read_file("MAPT.ZTM", &length);

register unsigned int idx = 0;
register int i=0;
register unsigned int ii=0;
register unsigned int iii=0;
register unsigned int nbpoint=0;
register unsigned int nbpolygon=0;
//register unsigned int MeshSize=0;
/*register POINT *pointArray;
register POLYGON *polygonArray;
register ATTR *attributeArray;*/

currentAddress = startAddress;

mesh_tot = jo_swap_endian_ushort(*((unsigned short *)(stream)));
idx+=2;

for (i=0; i<mesh_tot; i++)
{
    //MeshSize = jo_swap_endian_uint(*((unsigned int *)(stream+idx)));
    //idx+=4;
    jo3dMesh[i] = currentAddress;

    nbpoint = jo_swap_endian_uint(*((unsigned int *)(stream+idx)));
    idx+=4;
    nbpolygon = jo_swap_endian_uint(*((unsigned int *)(stream + idx)));
    idx+=4;

    //pointArray = (POINT*)currentAddress;
    jo3dMesh[i]->data.pntbl = (POINT*)(jo3dMesh[i] + sizeof(unsigned int));
    for (JO_ZERO(ii); ii<nbpoint; ii++)
    {
        for (JO_ZERO(iii); iii<3; iii++)
        {
            jo3dMesh[i]->data.pntbl[ii][iii] =  jo_swap_endian_int(*((int *)(stream + idx)));
            //pointArray[ii][iii] = jo_swap_endian_int(*((int *)(stream + idx)));
            idx+=4;
        }
    }
    jo3dMesh[i]->data.pltbl = (POLYGON*) (jo3dMesh[i]->data.pntbl + sizeof(POINT)*nbpoint);
    //polygonArray = (POLYGON*) (pointArray + sizeof (POINT)*nbpoint);
    for (JO_ZERO(ii); ii<nbpolygon; ii++)
    {
        for (JO_ZERO(iii); iii<3; iii++)
        {
            jo3dMesh[i]->data.pltbl[ii].norm[iii] = jo_swap_endian_int(*((int *)(stream + idx)));
            //polygonArray[ii].norm[iii]= jo_swap_endian_int(*((int *)(stream + idx)));
            idx+=4;
        }
        for (JO_ZERO(iii); iii<4; iii++)
        {
            jo3dMesh[i]->data.pltbl[ii].Vertices[iii]=jo_swap_endian_ushort(*((unsigned short*)(stream + idx)));
            //polygonArray[ii].Vertices[iii]=jo_swap_endian_ushort(*((unsigned short*)(stream + idx)));
            idx+=2;
        }
    }
    jo3dMesh[i]->data.attbl = (ATTR*) (jo3dMesh[i]->data.pltbl + sizeof(POLYGON)*nbpolygon);
   // attributeArray = (ATTR*) (polygonArray + sizeof(POLYGON)*nbpolygon);
    for (JO_ZERO(ii); ii<nbpolygon; ii++)
    {
        jo3dMesh[i]->data.attbl[ii].texno = jo_swap_endian_ushort(*((unsigned short*)(stream + idx)));
        idx+=2;
        jo3dMesh[i]->data.attbl[ii].sort = SORT_MAX;
        jo3dMesh[i]->data.attbl[ii].flag = Dual_Plane;
        jo3dMesh[i]->data.attbl[ii].colno = No_Palet;
        jo3dMesh[i]->data.attbl[ii].gstb = No_Gouraud;
        jo3dMesh[i]->data.attbl[ii].dir = MESHoff, sprNoflip, UseLight;
       /* attributeArray[ii].texno = jo_swap_endian_ushort(*((unsigned short*)(stream + idx)));
        attributeArray[ii].sort = SORT_MAX;
        attributeArray[ii].flag = Dual_Plane;
        attributeArray[ii].colno = No_Palet;
        attributeArray[ii].gstb = No_Gouraud;
        attributeArray[ii].dir = MESHoff, sprNoflip, UseLight;*/

    }

    jo3dMesh[i]->data.nbPoint = nbpoint;
    jo3dMesh[i]->data.nbPolygon = nbpolygon;
    /*jo3dMesh[i] = (jo_3d_mesh*) (attributeArray + sizeof(ATTR)*nbpolygon);
    jo3dMesh[i]->data.pntbl = pointArray;
    jo3dMesh[i]->data.pltbl = polygonArray;
    jo3dMesh[i]->data.nbPoint = nbpoint;
    jo3dMesh[i]->data.nbPolygon = nbpolygon;
    jo3dMesh[i]->data.attbl = attributeArray;*/

    currentAddress = (void*) (jo3dMesh[i] + sizeof(jo3dMesh[i]));
}
jo_free(stream);
jo_printf(0, 6, "current adress :            %d     ", currentAddress);
return currentAddress;
}


void jo_main(void)
{
	jo_core_init(JO_COLOR_Blue);
	jo_sprite_add_tga(JO_ROOT_DIR, "ROCK.TGA", JO_COLOR_Transparent);
	jo_3d_camera_init(&cam);
    jo_3d_window(0, 0, JO_TV_WIDTH - 1, JO_TV_HEIGHT - 1, 2500, JO_TV_WIDTH_2, JO_TV_HEIGHT_2);
	LoadBinaryMAP((void*)WORK_RAM_LOW);
	jo_core_add_callback(my_gamepad);
	jo_core_add_callback(my_draw);
	apply_effect();
	jo_core_run();
}

/*
 ** END OF FILE
 */
